#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

__doc__ = """Custom widgets for panels, forms and buttons in Vacca"""

import re,os,traceback,time,sys
import fandango,vacca
import fandango.functional as fun

from vacca.utils import * #Qt, Qwt5, etc
from taurus.qt.qtgui.taurusgui.utils import PanelDescription

import taurus
import taurus.qt
from taurus.qt.qtgui.container import TaurusWidget as WIDGET_CLASS
from taurus.qt.qtgui.panel import TaurusForm as FORM_CLASS
from taurus.qt.qtgui.panel.taurusdevicepanel import TaurusDevicePanel,get_regexp_dict,searchCl,matchCl,str_to_filter,get_White_palette,get_eqtype
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.panel.taurusform import TaurusForm,TaurusCommandsForm

try:
    #Taurus 4
    from taurus.qt.qtgui.icon import getCachedPixmap
    from taurus.core import TaurusAttribute,TaurusDevice,TaurusAuthority
except:
    #Taurus 3
    from taurus.qt.qtgui.resource import getPixmap as getCachedPixmap
    from taurus.core import TaurusAttribute,TaurusDevice,TaurusDatabase

###############################################################################
# Help Methods
def init_qt():
    return Qt.QApplication([])
            
def get_dev_attrs(dev):
    default = ['state','status']    
    #: DEV_ATTRS definition
    #: The lists of attributes to visualize for each vacuum device class    
    #attrs = taurus.Device(dev).get_attribute_list()
    for k,v in DEV_ATTRS.items():
        if re.match(k,dev.lower()): 
            return default+v
    return default
    
def get_eqtype(dev):
    ''' It extracts the eqtype from a device name like domain/family/eqtype-serial'''
    try: eq = str(dev).split('/')[-1].split('-',1)[0].upper()
    except: eq = ''
    return eq

###############################################################################
# Variables that control VaccaPanel shape
STATUS_HEIGHT=170
SPLIT_SIZES=[] #[15,50,35]

#APP_FOLDER = '/homelocal/sicilia/applications/vacca/'
EQUIP_IMAGE=wdir('image/equips/icon-%s.gif')
IMAGE_SIZE=(200,100) #(width,height)

class VaccaSplash(Qt.QSplashScreen, fandango.Singleton):

    def __init__(self, timeout=10000):
        self._pixmap = Qt.QPixmap(wdir('image/icons/alba-entry.JPG'))
        Qt.QSplashScreen.__init__(self, self._pixmap)
        self.showMessage('initializing application...')
        self._timer = Qt.QTimer.singleShot(timeout, self.close)
        self.show()

class DomainButton(Qt.QToolButton):

    def __init__(self, parent=None):
        #self.call__init__wo_kw(Qt.QToolButton, parent)
        Qt.QToolButton.__init__(self, parent)
        self.setLayout(Qt.QVBoxLayout())

        from taurus.qt.qtgui.display import TaurusLed
        self._led = TaurusLed()
        self._label = Qt.QLabel()
        self.layout().addWidget(self._led)
        self.layout().addWidget(self._label)
        self.layout().setAlignment(Qt.Qt.AlignCenter)

    def setLabel(self, txt):
        self._label.setText(txt)

    def setModel(self, model, action):
        #print 'DomainButton.setModel(%s,%s)'%(model,action)
        self._model = model = str(model)
        self._led.setModel(model)
        if not self._label.text(): self.setLabel(model.split('/')[0])
        self._cmd = action#action.split()
        self._action = taurus.qt.qtgui.util.ExternalAppAction(cmdargs=self._cmd)
        self.connect(self, Qt.SIGNAL("pressed()"), self.show)
        #_action.actionTriggered)

    def show(self, args=None):
        #print('VaccaAction(%s).show(%s,%s)'%(self._default_cmd,self._cmdargs,args))
        print 'Launching %s'%self._cmd
        self._action.actionTriggered(args)


class VaccaAction(Qt.QToolButton,taurus.qt.qtgui.base.TaurusBaseWidget):

    def __init__(self, parent=None, designMode=False, default=None):
        #print('VaccaAction()')
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QToolButton, parent)
        self.call__init__(taurus.qt.qtgui.base.TaurusBaseWidget, name,
                          designMode=designMode)
        self._text, self._icon, self._cmdargs, self._action = '', '', [], None
        self._default_cmd = default.split() if default is not None else []
        
    def setCommand(self, cmd):
        if not fandango.isSequence(cmd):
            cmd = fandango.toList(cmd)
        self._default_cmd = cmd

    def setModel(self, model):
        self.info('VaccaAction(%s).setModel(%s)' % (self._default_cmd, model))
        model = map(str,fun.toSequence(model))
        if len(model) > 1:
            self._text = model[0]
        if len(model) > 2:
            self._icon = model[1]
        self._cmdargs = model[2 if len(model) > 2 else 1 if len(model) > 1 else
        0:]
        self._action = taurus.qt.qtgui.util.ExternalAppAction(
            cmdargs=self._default_cmd+self._cmdargs,
            text=self._text,
            icon=self._icon
        )

        #self._action = ExternalApp(cmdargs=['mambo'], text="ArchivingViewer", icon=WDIR+'image/icons/Mambo-icon.png')
        self.setDefaultAction(self._action)
        self.setText(self._text)
        self.setToolButtonStyle(Qt.Qt.ToolButtonTextUnderIcon)

    def show(self, args=None):
        #print('VaccaAction(%s).show(%s,%s)'%(self._default_cmd,self._cmdargs,args))
        self._action.actionTriggered(args)

    def kill(self):
        if self._action:
            self._action.kill()
            
try:
  import MySQLdb
  from PyTangoArchiving.widget.ArchivingBrowser import ArchivingBrowser
  BROWSER = ArchivingBrowser
except:
  print('Failed to load PyTangoArchiving!')
  BROWSER = Qt.QWidget
  
class VaccaFinder(BROWSER):
  """
  Convenience class to deal with alternatives to ArchivingBrowser when not available in the PYTHONPATH
  """
  def __init__(self,parent=None,domains=None,regexp='*pnv-*',USE_SCROLL=True,USE_TREND=False):
        print 'In DomainChooser(), loading thermocouples from Tango'
        if BROWSER is Qt.QWidget:
            Qt.QWidget.__init__(self,parent)
        else:
            BROWSER.__init__(self,parent,domains,regexp,USE_SCROLL,USE_TREND)
  pass


class SimplePanel(WIDGET_CLASS):

    def getInnerPanel(self):
        return getattr(self, 'innerPanel', None)
        
    def setModel(self, model):
        if not hasattr(self, '_layout'):
            self._layout = Qt.QVBoxLayout()
            self.setLayout(self._layout)
            
        print('In SimplePanel.setModel(%s)' % model)
        prev = self.getInnerPanel()
        print('previous was %s,%s' % (prev, prev and prev.getModel()))
        old = prev and prev.getModel()
        if prev:
            try:
                print('\t dettaching %s' % old)
                self._layout.removeWidget(prev)
                #prev.setModel(None)
            except:
                traceback.print_exc()
        self.innerPanel = TaurusDevicePanel()
        self.innerPanel.setModel(model)
        self._layout.addWidget(self.innerPanel)
        
    def getModel(self):
        i = self.getInnerPanel()
        return i.getModel() if i else None

class VaccaDocker(Qt.QMainWindow):
    
    def __init__(self,*args):
        Qt.QMainWindow(self,*args)
        self.setDockNestingEnabled(True)
        
    @staticmethod
    def getDefaultIcon():
        """
        :return: The Default Icon Path.
        """
        path = 'image/widgets/DevicePanel.png'
        return path

    @staticmethod
    def getPanelDescription(name='VaccaDocker',model=''):
        """
        :param name: Name for the Panel
        :param model: Model for the panel
        :return:
        """
        return PanelDescription(name,'vacca.panel.VaccaDocker')
      
class VaccaSearchForm(TaurusWidget):
    """
    Form with an embedded search bar that can be used to set the models
    
    The preffix and suffix arguments can be used to limit the search
    """
  
    def __init__(self,preffix='',suffix='',labels=False,parent=None):
        TaurusWidget.__init__(self,parent)
        self.preffix = preffix
        self.suffix = suffix
        self.labels = labels
        if preffix or suffix:
          self.setWindowTitle('%s ... %s'%(preffix,suffix))
        self.setLayout(Qt.QVBoxLayout())
        self.bar = Qt.QWidget(self)
        self.search = Qt.QLineEdit()
        self.search.setText('... type to search for attributes ...')
        self.button = Qt.QPushButton('Load')
        self.connect(self.button,Qt.SIGNAL("pressed()"),self.apply_search)
        self.bar.setLayout(Qt.QHBoxLayout())
        self.bar.layout().addWidget(self.search)
        self.bar.layout().addWidget(self.button)
        self.layout().addWidget(self.bar)
        self.form = TaurusForm(self)
        self.layout().addWidget(self.form)
        
    def apply_search(self):
        signs = "==|=|<|>"
        txt = str(self.search.text() or '*')
        if fandango.clsearch(signs,txt):
          t = fandango.re.split(signs,txt,1)[0]
          txt,formula = t,txt.replace(t,'')
        else:
          formula = ''
        txt = txt.replace(' ','*').strip()
        txt = self.preffix+txt+self.suffix
        model = fandango.get_matching_attributes(txt)
        if self.labels and '/' in txt:
          dev,attr = txt.rsplit('/',1)
          model.extend(fandango.tango.get_matching_device_attribute_labels(dev,attr).keys())
        if formula and model:
          nm = []
          for m in model:
            try:
              f = '%s %s'%(fandango.read_attribute(m),formula)
              if eval(f):
                nm.append(m)
            except:
              pass
          model = nm
        self.form.setModel(model)
        
class VaccaPanel(fandango.qt.Dropable(taurus.qt.qtgui.panel.TaurusDevicePanel)):
    """

    :param parent:
    :param model: a device name or an state attribute
    :param palette:
    :param bound:
    :param filters:
    :return:

    It is a class that inherits from TaurusDevicePanel and Dropable module from fandango.

    If connect=True at init, it checks if it exists any shareDataManager to
        subscribe in it. It can be done later using getPanelDescription or connectSharedSignals

    This Widget shows the device Commands and Attributes, it is listening the shareDataManager to show the device selected information.

    The title of this Widget can be draggable.

    VaccaPanel has the follow functionalities:

        * Title is draggable.
        * Is connected to shareDataManager to share information in the GUI.

    """
    
    def __init__(self,parent=None,model=None, filters=[], connect=False): #,palette=None, bound=True,filters=[]):
        """
        In Init, the class VaccaPanel check if exist any shareDataManager to
        subscribe in it.

        :param parent:
        :param model: Model to start the Panel.
        :param filters: dictionary with 'attrs' and 'comms' filters as regexp or tuples lists
        :return:
        """
        
        self.call__init__(taurus.qt.qtgui.panel.TaurusDevicePanel, parent, model)
        #self.setLogLevel(self.Info)
        self.info('init(%s,%s): connecting ...'%(model,filters))
        
        self._connected = []
        if connect: self.connectSharedSignals()

        if self.checkDropSupport():
            self.setSupportedMimeTypes([self.TAURUS_DEV_MIME_TYPE,
                                        self.TEXT_MIME_TYPE, self.TAURUS_MODEL_MIME_TYPE])
            self.setDropEventCallback(self.setModelHook)

        self.info('init(...): layout ...')
        self._header.layout().removeWidget(self._label)
        
        self._label = fandango.qt.Dropable(fandango.qt.Draggable(Qt.QLabel))()
        self._label.font().setBold(True)
        self._label.setText('SELECT A DEVICE FROM THE TREE')
        self._header.layout().addWidget(self._label,0,1,Qt.Qt.AlignLeft)
        self._label.setDragEventCallback(self._label.text)
        self._label.checkDropSupport()
        self._label.setSupportedMimeTypes([self.TAURUS_DEV_MIME_TYPE,
                    self.TEXT_MIME_TYPE, self.TAURUS_MODEL_MIME_TYPE])
        self._label.setDropEventCallback(self.setModelHook)
        
        #self.setToolTip(getattr(self,'__help__',self.__doc__))
        
        if filters:
          self.info('VaccaPanel(filters=%s)'%filters)
          if 'attrs' in filters:
            type(self)._attribute_filter = {'.*':filters['attrs']}
          if 'comms' in filters:
            type(self)._command_filter = {'.*':
              [c if fun.isSequence(c) else (c,()) for c in filters['comms']]}
            
    def connectSharedSignals(self,read='SelectedInstrument',write=''):
        self.info('connectSharedSignals(%s,%s)'%(read,write))
        sdm = vacca.utils.get_shared_data_manager()
        if sdm and read and read not in self._connected:
            sdm.connectReader(read,self.setModelHook,readOnConnect=True) 
            self._connected.append(read)
        return
            
    def setModelHook(self,model):
        #self.info('%s,%s'%(repr(args),repr(kwargs)))
        #l = [(str(type(a)),) for l in (args,kwargs.values()) for a in args]
        self.info('In setModelHook(%s)'%str(model))
        try:
          fandango.tango.parse_tango_model(str(model).strip())['device']
          self.setModel(str(model))
        except:
          self.warning('Invalid model: %s\n%s'%(repr(model),traceback.format_exc()))
    
    @classmethod
    def getAttributeFilters(klass,dev_class=None):
        """
        TaurusDevicePanel filters are fixed to work by device name.
        But, if AttributeFilters is declared as class property, it will override them.
        
        get{Property}(klass,dev_class) will update it from Tango DB and return the matching values.
        set{Property}(dict) will update the parent TaurusDevicePanel dictionary, not the DB
        """
        if dev_class is not None and dev_class not in klass._attribute_filter:
            filters = get_class_property(dev_class,'AttributeFilters',extract=False)
            if filters:
                filters = [(l.split(':')[0],l.split(':')[-1].split(',')) for l in filters]
                klass._attribute_filter[dev_class] = filters
                #return {'.*':filters}
                
        print('getAttributeFilters(%s,%s): ...'%(klass,dev_class))#,klass._attribute_filter))
        return klass._attribute_filter
        
    @classmethod
    def getCommandFilters(klass,dev_class=None):
        """
        get{Property}(klass,dev_class) will update it from Tango DB and return the matching values.
        set{Property}(dict) will update the parent TaurusDevicePanel dictionary, not the DB
        """
        if dev_class is not None and dev_class not in klass._command_filter:
            filters = get_class_property(dev_class,'CommandFilters',extract=False)
            if filters:
                #filters = dict((k,eval(v)) for k,v in (l.split(':',1) for l in filters))
                filters = [(c,()) for c in filters]
                klass._command_filter[dev_class] = filters
                
        print('getCommandFilters(%s,%s): ...'%(klass,dev_class))#,klass._command_filter))
        return klass._command_filter
      
    @classmethod
    def getIconMap(klass,dev_class=None):
        if dev_class is not None and dev_class not in klass._icon_map:
            p = get_class_property(dev_class,'Icon',extract=True)
            if p: klass._icon_map[dev_class] = p #Not trivial!
        print('getIconMap(%s): ...'%(klass))
        #print('getIconMap(%s): ...%s'%(klass,klass._icon_map))
        return klass._icon_map
      
    def setModel(self,model,pixmap=None):
        """
        Set Model is the callback used in shareDataManager to manage device
        selections.

        :param model: Model to VaccaPanel
        :param pixmap:
        :return:
        """
        try:    
          #self.setLogLevel(self.Debug)
          self.info('VaccaPanel(%s).setModel(%s,%s)'%(id(self),model,pixmap))
          self.checkDropSupport() #<< Needed to reapply drop support overriden by taurusgui.ini
          model,modelclass,raw = str(model).strip(),'',model
          model = fandango.tango.parse_tango_model(model)
          if model is None: 
            self.warning('VaccaPanel(%s).setModel(%s,%s): MODEL NOT PARSABLE!'%(id(self),model,pixmap))
            return
          else: model = model['device']
          if model: 
            model = model and model.split()[0] or ''
            modelclass = taurus.Factory().findObjectClass(model)
          self.debug('In TaurusDevicePanel.setModel(%s(%s),%s)'%(raw,modelclass,pixmap))
          if model == self.getModel():
            return
          elif raw is None or not model or not modelclass: 
            if self.getModel(): self.detach()
            return
          elif issubclass(modelclass, TaurusAttribute):
            #if model.lower().endswith('/state'): 
            model = model.rsplit('/',1)[0]
          elif not issubclass(modelclass, TaurusDevice):
            self.warning('TaurusDevicePanel accepts only Device models')
            return
        except Exception,e:
          traceback.print_exc()
          raise e
          
        try:
            try: taurus.Device(model).getDeviceProxy().ping() #T4
            except: taurus.Device(model).getHWObj().ping() #T3
            dev_class = fandango.get_device_info(model).dev_class
            if self.getModel(): self.detach() #Do not dettach previous model before pinging the new one (fail message will be shown at except: clause)
            TaurusWidget.setModel(self,model)
            self.setWindowTitle(str(model).upper())
            model = self.getModel()
            self._label.setText(model.upper())
            font = self._label.font()
            font.setPointSize(15)
            self._label.setFont(font)
            if pixmap is None and self.getIconMap(dev_class=dev_class):
                for k,v in self.getIconMap().items():
                    if searchCl(k,model) or searchCl(k,dev_class):
                        pixmap = v
                        break
            if pixmap is not None:
                #print 'Pixmap is %s'%pixmap
                qpixmap = Qt.QPixmap(pixmap)
                if qpixmap.height()>.9*IMAGE_SIZE[1]: qpixmap=qpixmap.scaledToHeight(.9*IMAGE_SIZE[1])
                if qpixmap.width()>.9*IMAGE_SIZE[0]: qpixmap=qpixmap.scaledToWidth(.9*IMAGE_SIZE[0])
            else:
                qpixmap = getCachedPixmap(':/logo.png')
            
            self._image.setPixmap(qpixmap)
            self._state.setModel(model+'/state')
            if hasattr(self,'_statelabel'): self._statelabel.setModel(model+'/state')
            self._status.setModel(model+'/status')
            
            try:
                self._attrsframe.clear()
                class_filters = type(self).getAttributeFilters(dev_class=dev_class)
                filters = get_regexp_dict(self._attribute_filter,model,[])
                if not filters: 
                  filters = get_regexp_dict(self._attribute_filter,dev_class,['.*'])
                  
                search_tab = None
                if filters == ['.*'] and len(taurus.Device(model).get_attribute_list())>32:
                    filters = [
                      #('Status',['Status']),
                      ('Search Attributes',['.*'])
                      ]
                
                if hasattr(filters,'keys'): filters = filters.items() #Dictionary!
                print('\tfilters = %s'%filters)

                #Showing multiple Tabs
                if filters and isinstance(filters[0],(list,tuple)): 
                    self._attrs = []
                    for tab,attrs in filters:
                      if attrs[1:] or attrs[0] not in ('*','.*'):
                        self.debug('VaccaPanel.setModel.get_attrs_form(%s,%s)'%(model,attrs))
                        self._attrs.append(self.get_attrs_form(device=model,filters=attrs,parent=self))
                        self._attrsframe.addTab(self._attrs[-1],tab)
                      else:
                        self.info('Embedding a Search panel')
                        search_tab = tab,VaccaSearchForm(preffix=model+'/*',suffix='*',labels=True)
                #Mapping into a single form
                else: 
                    if self._attrs and isinstance(self._attrs,list): self._attrs = self._attrs[0]
                    if not isinstance(self._attrs,Qt.QWidget): self._attrs = None
                    self._attrs = self.get_attrs_form(device=model,form=self._attrs,filters=filters,parent=self)
                    if self._attrs: self._attrsframe.addTab(self._attrs,'Attributes')
                    
                if not TaurusDevicePanel.READ_ONLY:
                    self._comms = self.get_comms_form(model,self._comms,self)
                    if self._comms: self._attrsframe.addTab(self._comms,'Commands')
                    
                    if search_tab:
                      self._attrsframe.addTab(search_tab[1],search_tab[0])
                      
                if SPLIT_SIZES: self._splitter.setSizes(SPLIT_SIZES)
                
            except Exception,e:
                self.warning('setModel(%s) failed!'%model)
                self.warning( traceback.format_exc())
                qmsg = Qt.QMessageBox(Qt.QMessageBox.Critical,'%s Error'%model,'%s not available'%model,Qt.QMessageBox.Ok,self)
                qmsg.setDetailedText(traceback.format_exc())
                qmsg.show()
                raise e

        except Exception,e:
            self.warning('setModel(%s) failed!'%model)
            self.warning(traceback.format_exc())
            qmsg = Qt.QMessageBox(Qt.QMessageBox.Critical,'%s Error'%model,'%s not available'%model,Qt.QMessageBox.Ok,self)
            qmsg.setDetailedText(traceback.format_exc())
            qmsg.show()
            raise e
            
        self.setWindowTitle(self.getModel())
        return
    
    def get_comms_form(self,device,form=None,parent=None):
      
        self.trace( 'In TaurusDevicePanel.get_comms_form(%s)'%device)
        dev_class = fandango.get_device_info(device).dev_class
        filters = type(self).getCommandFilters(dev_class)
        params = get_regexp_dict(filters,device,[]) or get_regexp_dict(filters,dev_class,[])
        
        if filters and not params: #If filters are defined only listed devices will show commands
            self.debug('TaurusDevicePanel.get_comms_form(%s): By default an unknown device type will display no commands'% device)
            return None 
          
        if not form: 
            form = TaurusCommandsForm(parent)
            
        elif hasattr(form,'setModel'): 
            form.setModel('')
            
        try:
            form.setModel(device)
            if params: 
                form.setSortKey(lambda x,vals=[s[0].lower() for s in params]: vals.index(x.cmd_name.lower()) if str(x.cmd_name).lower() in vals else 100)
                form.setViewFilters([lambda c: str(c.cmd_name).lower() not in ('state','status') and any(searchCl(s[0],str(c.cmd_name)) for s in params)])
                form.setDefaultParameters(dict((k,v) for k,v in (params if not hasattr(params,'items') else params.items()) if v))
            for wid in form._cmdWidgets:
                if not hasattr(wid,'getCommand') or not hasattr(wid,'setDangerMessage'): continue
                if re.match('.*(on|off|init|open|close).*',str(wid.getCommand().lower())):
                    wid.setDangerMessage('This action may affect other systems!')
            #form._splitter.setStretchFactor(1,70)
            #form._splitter.setStretchFactor(0,30)
            form._splitter.setSizes([80,20])
            
        except Exception: 
            self.warning('Unable to setModel for TaurusDevicePanel.comms_form!!: %s'%traceback.format_exc())
        return form

    @staticmethod
    def getDefaultIcon():
        """
        :return: The Default Icon Path.
        """
        path = vpath('image/widgets/DevicePanel.png')
        return path

    @staticmethod
    def getPanelDescription(name='VaccaDevicePanel',model=''):
        """
        :param name: Name for the Panel
        :param model: Model for the panel
        :return:
        """
        return PanelDescription(
          name,classname='vacca.panel.VaccaPanel',
          model=model,sharedDataRead={'SelectedInstrument':'setModelHook'},
          )

        
def configure_form(dev,form=None):
    """ Creates a TauForm and configures its Status fields 
    """
    if form is None:
        form = FORM_CLASS()
    elif hasattr(form, 'setModel'):
        form.setModel([])
    ##Configuring the TauForm:
    form.setWithButtons(False)
    form.setWindowTitle(dev)
    form.setModel('%s/%s' % (dev, a) for a in get_dev_attrs(dev))
    ##Adapting the status widget to show properly an status
    status = form.getItemByModel(dev+'/status') #A TauValue object (containing label and value widgets)
    sw = status.readWidget() #A TauStatusLabel object
    sw.setAlignment(Qt.Qt.AlignLeft)
    sw.setMinimumHeight(STATUS_HEIGHT)
    #sw.setShowQuality(False) #It didn't work as expected
    return form


def main(args):

    import sys,re,traceback,taurus
    assert len(args)>1, '\n\nUsage:\n\t> python panel.py [a/device/name or synoptic.jdw] [--attrs] [attr_regexps] --comms [comm_regexps]'
    
    model = args[1]
    taurus.setLogLevel(taurus.core.util.Logger.Debug)
    app = Qt.QApplication(args)
    form = None
    
    if re.match('[\w-]+/[\w-]+/[\w-]+.*',model):
      
        print 'loading a device panel'
        k,filters = '--attrs',fandango.defaultdict(list)
        for f in args[2:]:
          if f.startswith('--'): k = f.strip('-')
          else: filters[k].append(f) #(f,()) if k=='comms' else f)
          
        form = VaccaPanel(filters=filters)  #palette=get_fullWhite_palette()
        form.setModel(model)
        
    elif model.lower().endswith('.jdw'):
        print 'loading a synoptic'
        form = taurus.qt.qtgui.graphic.TauJDrawSynopticsView(designMode=False,
          updateMode=taurus.qt.qtgui.graphic.TauJDrawSynopticsView.NoViewportUpdate
          ) 
        #FullViewportUpdate, : AWFUL CPU USAGE!!!!!!!!
        #MinimalViewportUpdate, : Qt Defaults
        #SmartViewportUpdate, : ?
        #BoundingRectViewportUpdate, : ?
        #NoViewportUpdate : Tau defaults
        form.setModel(model)
        models = form.get_item_list()
        for m in models:
            m = str(m)
            if m.count('/')==2: m+='/state'
            period = 120000.
            try: 
                taurus.Attribute(m).changePollingPeriod(period)
            except: print '(%s).changePollingPeriod(%s): Failed: %s'%(m,period,traceback.format_exc())
    
    print 'showing ...'
    form.show()
    sys.exit(app.exec_())

__doc__ = vacca.get_autodoc(__name__,vars())

if __name__ == '__main__': 
  main(sys.argv)



#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import sys,re,traceback
import fandango,fandango.qt,taurus
from taurus.external.qt import Qt,Qwt5
from taurus.qt.qtgui.taurusgui.utils import PanelDescription
from taurus.qt.qtcore.mimetypes import TAURUS_ATTR_MIME_TYPE, \
    TAURUS_DEV_MIME_TYPE, TAURUS_MODEL_MIME_TYPE, TAURUS_MODEL_LIST_MIME_TYPE
from taurus import Manager

from fandango import partial,FakeLogger as FL

from taurus.qt.qtgui.graphic import TaurusJDrawSynopticsView,TaurusGraphicsScene,TaurusJDrawGraphicsFactory,TYPE_TO_GRAPHICS
from taurus.qt.qtgui.graphic.taurusgraphic import TaurusGroupItem,TaurusGraphicsItem,QGraphicsTextBoxing
from taurus.qt.qtgui.util import ExternalAppAction

from vacca.utils import get_env_variable
from vacca.panel import VaccaPanel

try:
    #Taurus 3
    from taurus.core.taurusvalidator import DeviceNameValidator as TaurusDeviceNameValidator, AttributeNameValidator as TaurusAttributeNameValidator
except:
    #Taurus 4
    from taurus.core.taurusvalidator import TaurusDeviceNameValidator, TaurusAttributeNameValidator

import traceback
def catched(f):
  def new_f(*args,**kwargs):
    try:
      return f(*args,**kwargs)
    except:
      traceback.print_exc()
  return new_f

__doc__ = """
How Synoptics work in Taurus
============================

Parsing of JDRAW files
----------------------

The ...graphic.jdraw.jdraw module provides the TaurusJDrawGraphicsFactory
At graphic.jdraw_view.TaurusJDrawSynopticsView.setModel(), the factory is passed
to the jdraw_parser.parse method. It will return a TaurusGraphicsScene object, 
that will be linked to the view using ...View.setScene() method.

Once the Scene is set, the View object is just a proxy to the scene. 
Its main function is to manage resizing, signals and drag&drop.

The Scene provides all the methods for selection and interaction with objects.

In Vacca the View, Factory and Scene objects are overriden by:

- VaccaSynoptic(TaurusJDrawSynopticsView)
- VaccaSynopticGraphicsFactory(TaurusJDrawGraphicsFactory)
- VaccaGraphicsScene(TaurusGraphicsScene)

Parsing of SVG files
--------------------

It will require the maxIV plugin. It provides a javascript code to be executed
within a QWeb widget. Qt events are used to pass the states to javascript.

"""

class QSignalHook(Qt.QObject):
    """
    @TODO: This class belongs to Fandango!!, remove it once fandango will be updated with the str(model) fix
    Class initialized with a transformation function.
    For every setModel(model) it will emit a modelChanged(function(model))
    """
    #__pyqtSignals__ = ("modelChanged",)
    def __init__(self,function):
        Qt.QObject.__init__(self)
        self._model = None
        self._new_model = None
        self.function = function
        
    def setModel(self,model):
        model = str(model) #Do a copy or die!
        from fandango import FakeLogger as FL
        fl = FL('QSignalHook',True)
        try:
          #fl.warning('QSignalHook(%s)'%(self._model,))
          self._model = model
          self._new_model = self.function(model) if self.function is not None else model
          #fl.warning('QSignalHook.modelChanged(%s => %s)'%(self._model,self._new_model))
          self.emit(Qt.SIGNAL("modelChanged"), self._new_model)
        except:
          fl.warning('QSignalHook: %s'%(traceback.format_exc()))

class VaccaSynoptic(TaurusJDrawSynopticsView):
    """

    Overrides TaurusJDrawSynopticsView methods to highlight multiple items in multiple colors.
    
    It allows to setup a delayed setModel call to reduce "hungs" on heavy synoptics.
    But!: This delayed call may cause problems in some versions of taurusGUI. (due to scene not-set yet)
    """

    JDRAW_FILE = None
    JDRAW_HOOK = None
    JDRAW_TREE = None

    
    def __init__(self, parent = None, delay=0, designMode = False, updateMode=None, alias = None, resizable = True, panelClass = VaccaPanel):
        self.delay = delay
        TaurusJDrawSynopticsView.__init__(self, parent,designMode,updateMode,alias,resizable,panelClass)
        #self.setLogLevel(self.Debug) #getattr(self,get_env_variable('VACCA_LOG') or 'Info',self.Info))
        self.setModelInConfig(False)

    
    def setModel(self, model):
        if self.delay:
            self.warning('VaccaSynoptic.setModel() delayed %d ms'%self.delay)
            self._setter = Qt.QTimer.singleShot(
                self.delay, 
                partial(self._setJDrawModel, model))
            self.JDRAW_FILE = model
        else:
            TaurusJDrawSynopticsView.setModel(self, model)

    
    def _setJDrawModel(self, model):
        print '>'*80
        print 'setJDrawModel(%s)' % model
        TaurusJDrawSynopticsView.setModel(self, model)

    
    def getModelMimeData(self):
        """ Used for drag events """

        model,mimeData = '',None
        try:
            #model = getattr(self.scene().itemAt(*self.mousePos),'_name','')
            selected = self.scene()._selectedItems
            if not selected:
                self.debug('jdrawView.getModelMimeData(%s): nothing to drag'%model)
                return
            model = getattr(([s for s in selected if getattr(s,'_name','')] or [selected])[0],'_name','')
            self.debug('getModelMimeData(%s)'%model)
            mimeData = Qt.QMimeData()
            if model:
                # self.debug('getMimeData(): DeviceModel at %s: %s',self.mousePos,model)
                mimeData.setText(model)
                mimeData.setData(TAURUS_MODEL_MIME_TYPE, model)
                if TaurusDeviceNameValidator().getParams(model):
                    self.debug('getMimeData(): DeviceModel at %s: %s',self.mousePos,model)
                    mimeData.setData(TAURUS_DEV_MIME_TYPE,model)
                elif TaurusAttributeNameValidator().getParams(model):
                    self.debug('getMimeData(): AttributeModel at %s: %s',self.mousePos,model)
                    mimeData.setData(TAURUS_ATTR_MIME_TYPE,model)
                    mimeData.setData(TAURUS_DEV_MIME_TYPE,model.rsplit('/',1)[0])
                else:
                    self.debug('getMimeData(): UnknownModel at %s: %s',self.mousePos,model)
        except:
            self.debug('jdrawView.getModelMimeData(%s): unable to get MimeData'%model)
            self.debug(traceback.format_exc())
        return mimeData
        
    @staticmethod
    def getPanelDescription(NAME='Synoptic',JDRAW_FILE='',JDRAW_HOOK=None,JDRAW_TREE=[]):
        """

        :param NAME: Name for the Panel
        :param JDRAW_FILE: synoptic configuration file
        :param JDRAW_HOOK: ''
        :param JDRAW_TREE: None
        :return: PanelDescription Object
        """

        # CONNECTING FAKELOGGER FOR DEBUGGING
        try:
            import vacca
            sdm = vacca.utils.get_shared_data_manager()
            if sdm:
                v = sdm._SharedDataManager__models.get('SelectedInstrument')
                # An inline instantiated FakeLogger will print selected instrument on each click
                sdm.connectReader('SelectedInstrument', #<<< Works when triggered from tree
                    FL('SDM.SelectedInstrument [%s,%s]' % 
                       (v.readerCount(),v.writerCount()), True).info,
                    readOnConnect=False)
        except:
            print '#'*80
            print('Shared Data Manager is not available! (no TaurusGUI instance?)')
            traceback.print_exc()
            sdm = None
            print '#'*80


        print '>'*20+'Loading Synoptic panel new ... %s, %s, %s'%\
            (JDRAW_FILE,JDRAW_HOOK,JDRAW_TREE)

        class_name,rsignal,wsignal = '',{},{}            
        if not JDRAW_FILE.endswith('.svg'): #Assuming a JDraw file
            print '>'*20+'Creating JDW VaccaSynoptic'

            if JDRAW_HOOK is not None:
                print 'Enabling JDRAW_HOOK = %s'%JDRAW_HOOK
                VaccaSynoptic.in_hook = QSignalHook(JDRAW_HOOK) #MUST BE PERSISTENT
                VaccaSynoptic.out_hook = QSignalHook(JDRAW_HOOK)

                #Synoptic will write this signal
                wsignal = {'JDrawOut': 'graphicItemSelected(QString)', }

                if sdm:
                    sdm.connectWriter('SelectedInstrument', VaccaSynoptic.out_hook,
                                      'modelChanged')

                    sdm.connectReader('JDrawOut', VaccaSynoptic.out_hook.setModel, #ITS NOT CALLED!?!?!?
                                      readOnConnect=False)

                    sdm.connectReader('JDrawOut', FL('SDM.JDrawOut',True).info,
                                      readOnConnect=False)
                    
                    v = sdm._SharedDataManager__models.get('JDrawOut')

                    sdm.connectReader('JDrawOut', FL('SDM.JDrawOut DONE [%s,'
                                                     '%s]'%(v.readerCount(),
                                                            v.writerCount()),
                                                     True).info,
                                      readOnConnect=False)

                    sdm.connectWriter('JDrawIn', VaccaSynoptic.in_hook,'modelChanged')
                    sdm.connectReader('JDrawIn', FL('SDM.JDrawIn', True).info,
                                      readOnConnect=False)
                    rsignal = {'JDrawIn': 'selectGraphicItem', }
                    sdm.connectReader('SelectedInstrument', VaccaSynoptic.in_hook.setModel,
                                      readOnConnect=False)

                    v = sdm._SharedDataManager__models.get('JDrawIn')
                    sdm.connectReader('JDrawIn', FL('SDM.JDrawIn DONE [%s,'
                                                    '%s]' % (v.readerCount(),
                                                           v.writerCount()),
                                                    True).info,
                                      readOnConnect=False)
            else:
                rsignal = {'SelectedInstrument': 'selectGraphicItem'}
                rsignal['HighlightInstruments'] = 'setHighlightedItems'
                wsignal = {'SelectedInstrument': 'graphicItemSelected(QString)'}

            if JDRAW_TREE:
                wsignal['LoadItems'] = 'modelsChanged'


            class_name='vacca.synoptics.VaccaSynoptic'

        elif JDRAW_FILE.endswith('.svg'):
            try:
                from svgsynoptic import SynopticWidget, Registry
                rsignal = {'SelectedInstrument': 'select_devices'}
                #@TODO: rsignal['HighlightInstruments'] = 'setHighlightedItems'
                wsignal = {'SelectedInstrument': 'graphicItemSelected(QString)'}
                class_name='synoptic.SynopticWidget'
            except:
                print('Horreur!: svgsynoptic MODULE NOT AVAILABLE!')
                return None

        print 'Out of VaccaSynoptic.getPanelDescription(%s,%s)'%(class_name,JDRAW_FILE)
        return PanelDescription(NAME,
                                #classname = 'vacca.VacuumSynoptic',
                                classname=class_name,
                                model=JDRAW_FILE, #Model loading is delayed by
                                # VacuumSynoptic method
                                sharedDataRead=rsignal,
                                sharedDataWrite=wsignal,
                                )        

    ###########################################################################
    # OVERLOADED METHODS FOR CUSTOM HIGHLITHING
    # It should be included in Taurus as patches
    
    
    def setHighlightedItems(self, models = [], color=Qt.Qt.red):
        """

        :param models: List of model to be Highlighted
        :param color: Color, by default Qt.Qt.red
        """
        if fandango.isSequence(models):
            models = '(%s)'%')|('.join(models)
        self.info('setHighLightedItems(%s)' % models)
        try:
            default = self.scene().selectionColor()
            self.scene().setSelectionColor(color)
            self.scene().selectGraphicItem(models)
            self.scene().setSelectionColor(default)
        except:
            err = traceback.format_exc()
            self.error(err)
        finally:
            self.scene().setSelectionColor(default)
    
    
    def getGraphicsFactory(self,delayed=False):
        return VaccaSynopticGraphicsFactory(self,alias=(self.alias or None),delayed=delayed)
    
    @staticmethod
    def getDefaultIcon():
        """
        :return: The Default Icon Path.
        """
        path = 'image/icons/Synoptic.png'
        return path

#A decorator for QGraphics Objects
def GetClassWithExtensions(klass):
  
    class ClassWithExtensions(klass):
        
        def setName(self,name):
            name = str(name or '') #self.__class__.__name__) #Using class name corrupted element selection
            self._name = name#srubio@cells.es: modified to store ._name since initialization (even if a model is not set)
        
        def getExtensions(self):
            """ Any in ExtensionsList, noPrompt, standAlone, noTooltip, noSelect, 
            ignoreRepaint, shellCommand, className, classParams """
            self._extensions = getattr(self,'_extensions',{})
            if 'ExtensionsList' in self._extensions:
                self._extensions.update((k.strip(),True) for k in self._extensions['ExtensionsList'].split(','))
                self._extensions.pop('ExtensionsList')
            for k in ('noPrompt','standAlone','noTooltip','ignoreRepaint','noSelect'):
                if self._extensions.get(k,None)=='': self._extensions[k] = True
            self.noPrompt = self._extensions.get('noPrompt',False)
            self.standAlone = self._extensions.get('standAlone',False)
            self.noTooltip = self._extensions.get('noTooltip',False)
            self.ignoreRepaint = self._extensions.get('ignoreRepaint', getattr(self,'ignoreRepaint',False))
            self.setName(self._extensions.get('name',self._name))
            tooltip = '' if (self.noTooltip or self._name==self.__class__.__name__ or self._name is None) else str(self._name)
            #self.debug('setting %s.tooltip = %s'%(self._name,tooltip))
            self.setToolTip(tooltip)
            #self.debug('%s.getExtensions(): %s'%(self._name,self._extensions))
            return self._extensions
          
    return ClassWithExtensions

class VaccaSynopticGraphicsFactory(TaurusJDrawGraphicsFactory):

    
    def getSceneObj(self,items):
        scene = VaccaGraphicsScene(self.myparent)
        for item in items:
            try:
                if isinstance(item, Qt.QWidget):
                    scene.addWidget(item)
                elif isinstance(item, Qt.QGraphicsItem):
                    scene.addItem(item)
            except:
                self.warning("Unable to add item %s to scene" % str(item))
                self.debug("Details:", exc_info=1)
        return scene
    
    
    def getGraphicsItem(self,type_,params):
        name = params.get(self.getNameParam())
        #applying alias
        for k,v in getattr(self,'alias',{}).items():
            if k in name:
                name = str(name).replace(k,v)
                params[self.getNameParam()] = name
        cls = None
        if '/' in name:
            #replacing Taco identifiers in %s'%name
            if name.lower().startswith('tango:') and (name.count('/')==2 or not 'tango:/' in name.lower()): 
                nname = name.split(':',1)[-1]
                params[self.getNameParam()] = name = nname
            if name.lower().endswith('/state'): name = name.rsplit('/',1)[0]
            cls = Manager().findObjectClass(name)
        else: 
            if name: self.debug('%s does not match a tango name'%name)
        klass = self.getGraphicsClassItem(cls, type_)
        
        self.debug(str((cls,type_,klass,klass.__name__)))
        if not hasattr(klass,'getExtensions'):
            klass = GetClassWithExtensions(klass)
            
        item = klass()
        ## It's here were Attributes are subscribed
        self.set_common_params(item,params)
        if hasattr(item,'getExtensions'):
            item.getExtensions() #<= must be called here to take extensions from params
            
        return item    
    
class VaccaGraphicsScene(TaurusGraphicsScene):
        
    
    def setSelectionColor(self,color):
        self._selectioncolor = color
        
    
    def selectionColor(self):
        try:
            assert self._selectioncolor
        except:
            self._selectioncolor = Qt.Qt.blue
        return self._selectioncolor
    
    def _displaySelectionAsOutline(self, items):
        def _outline(shapes):
            """"Compute the boolean union from a list of QGraphicsItem. """
            shape = None
            # TODO we can use a stack instead of recursivity
            for s in shapes:
                # TODO we should skip text and things like that
                if isinstance(s, TaurusGroupItem):
                    s = _outline(s.childItems())
                    if s == None:
                        continue

                s = s.shape()
                if shape != None:
                    shape = shape.united(s)
                else:
                    shape = s

            if shape == None:
                return None

            return Qt.QGraphicsPathItem(shape)

        # TODO we can cache the outline instead of computing it again and again
        selectionShape = _outline(items)
        if selectionShape:
            # copy-paste from getSelectionMark
            color = Qt.QColor(self.selectionColor())
            color.setAlphaF(.10)
            pen = Qt.QPen(Qt.Qt.SolidLine)
            pen.setWidth(4)
            pen.setColor(Qt.QColor(self.selectionColor()))
            selectionShape.setBrush(color)
            selectionShape.setPen(pen)

            for item in items:
                if item not in self._selectedItems: self._selectedItems.append(item)

            # TODO i dont think this function work... or i dont know how...
            #self.setSelectionMark(picture=selectionShape)
            # ... Then do it it with hands...
            # copy-paste from drawSelectionMark
            self._selection.append(selectionShape)
            # It's better to add it hidden to avoid resizings
            selectionShape.hide()
            self.addItem(selectionShape)
            # Put on Top
            selectionShape.setZValue(9999)
            selectionShape.show()
            self.updateSceneViews()

            return True

        return False
    
    def getSelectionMark(self,picture=None,w=10,h=10):
        if picture is None:
            if self.SelectionMark:
                SelectionMark = self.SelectionMark()
            else:
                SelectionMark = Qt.QGraphicsEllipseItem()
                color = Qt.QColor(self.selectionColor())
                color.setAlphaF(.10)
                SelectionMark.setBrush(color)
                pen = Qt.QPen(Qt.Qt.CustomDashLine)
                pen.setWidth(4)
                pen.setColor(Qt.QColor(self.selectionColor()))
                SelectionMark.setPen(pen)
                SelectionMark.hide() #It's better to add it hidden to avoid resizings                
        else:
            try:
                if isinstance(picture,Qt.QGraphicsItem):
                    SelectionMark = picture
                    SelectionMark.setRect(0,0,w,h)
                    SelectionMark.hide()
                elif operator.isCallable(picture):
                    SelectionMark = picture()
                else:
                    if isinstance(picture,Qt.QPixmap):
                        pixmap = picture
                    elif isinstance(picture,basestring) or isinstance(picture,Qt.QString):
                        picture = str(picture)
                        pixmap = Qt.QPixmap(os.path.realpath(picture))
                    SelectionMark = Qt.QGraphicsPixmapItem()
                    SelectionMark.setPixmap(pixmap.scaled(w,h))
                    SelectionMark.hide()
            except:
                self.debug('In setSelectionMark(%s): %s'%(picture,traceback.format_exc()))
                picture = None
        return SelectionMark
      
    ##################################################################
    # METHODS JUST FOR DEBUGGING
      
    def mousePressEvent(self,mouseEvent):
        #self.logger.setLogLevel(self.logger.Debug)
        self.debug('<'*80)
        self.debug('In TaurusGraphicsScene.mousePressEvent(%s,%s))'%(
          str(type(mouseEvent)),str(mouseEvent.button())))
        
        try: 
            obj = self.getItemClicked(mouseEvent)
            obj_name = getattr(obj,'_name', '')
            
            ##@BUG:OBJ NAME SHOULD BE KEPT EMPTY; THIS LINE CAUSED SOME SELECTIONS TO FAIL!
            if not obj_name and isinstance(obj,QGraphicsTextBoxing): obj_name = obj.toPlainText()
            
            if (mouseEvent.button() == Qt.Qt.LeftButton):
                """ A null obj_name should deselect all, we don't send obj because 
                we want all similar to be matched  """
                if self.selectGraphicItem(obj_name):
                    self.debug(' => graphicItemSelected(QString)(%s)'%obj_name)
                    self.emit(Qt.SIGNAL("graphicItemSelected(QString)"),obj_name)
                else:
                    # It should send None but the signature do not allow it
                    self.emit(Qt.SIGNAL("graphicItemSelected(QString)"), "")
                    
            if (mouseEvent.button() == Qt.Qt.RightButton):
                ''' This function is called when right clicking on TaurusDevTree area. 
                A pop up menu will be shown with the available options. '''
                self.debug('RightButton Mouse Event on %s, (%s)'%(
                  obj_name,isinstance(obj,TaurusGraphicsItem)))
                
                if isinstance(obj,TaurusGraphicsItem) and (obj_name or obj.contextMenu() or obj.getExtensions()):
                    self.showContextMenu(obj,mouseEvent)

        except Exception:
            self.warning(traceback.format_exc())
        self.debug('<'*80)
            
    def showContextMenu(self,obj,mouseEvent):
        obj_name = getattr(obj,'_name', '')
        self.debug('%s.showContextMenu()'%(obj_name or obj))

        def addMenuAction(menu,k,action,last_was_separator=False):
            self.debug('%s.addMenuAction(%s,%s,%s)'%(obj_name or obj,menu,k,action))
            try:
                if k:
                    configDialogAction = menu.addAction(k)
                    if action: 
                        self.connect(configDialogAction, Qt.SIGNAL("triggered()"), 
                                    lambda dev=obj_name,act=action: act(dev))
                    else: configDialogAction.setEnabled(False)
                    last_was_separator = False
                elif not last_was_separator: 
                    menu.addSeparator()
                    last_was_separator = True
            except Exception,e: 
                self.warning('Unable to add Menu Action: %s:%s'%(k,e))                    
            return last_was_separator
        
        menu = Qt.QMenu(None)#self.parent)    
        last_was_separator = False
        extensions = obj.getExtensions()
        
        self.debug('%s.showContextMenu(%s,%s)'%(obj_name or obj,extensions,obj.contextMenu()))
        
        if obj_name and (not extensions or not extensions.get('className')): 
            #menu.addAction(obj_name)
            addMenuAction(menu,'Show %s panel'%obj_name,lambda x=obj_name: self.showNewPanel(x))
            
        if obj.contextMenu():
          
            if obj_name: 
                menu.addSeparator()
                last_was_separator = True
                
            for t in obj.contextMenu(): #It must be a list of tuples (ActionName,ActionMethod)
                last_was_separator = addMenuAction(menu,t[0],t[1],last_was_separator)
                
        if extensions:
            if not menu.isEmpty(): menu.addSeparator()
            className = extensions.get('className')
            if className and className!='noPanel':
                self.debug('launching className extension object')
                addMenuAction(menu,'Show %s'%className,lambda d,x=obj: self.showNewPanel(x))
                
            if extensions.get('shellCommand'):
                addMenuAction(menu,'Execute',lambda d,x=obj: self.getShellCommand(x))
                
        if not menu.isEmpty():
            self.debug('%s(%s).menu.exec()'%(obj,obj_name))
            menu.exec_(Qt.QPoint(mouseEvent.screenPos().x(),mouseEvent.screenPos().y()))
            
        del menu
      
    ##################################################################
    

###############################################################################

import doc
__doc__ = doc.get_autodoc(__name__,vars())

def test(model,filters='',debug=False):
    
    if debug: 
        taurus.setLogLevel(taurus.core.util.Logger.Debug)
        form.setWindowTitle(model)
    print 'loading synoptic: %s'%model
    form = VaccaSynoptic(delay=1000,designMode=False)
    #form = taurus.qt.qtgui.graphic.TaurusJDrawSynopticsView(designMode=False)
    #designMode=False,updateMode=VaccaSynoptic.NoViewportUpdate)
    form.show()
    form.setModel(model)
    form.setWindowTitle(model)
    print 'showing ...'
    return form

if __name__ == '__main__':
    #!/usr/bin/python
    assert len(sys.argv)>1, '\n\nUsage:\n\t> python synoptic [jdw file]'
    
    app = Qt.QApplication([]) #sys.argv)

    model = sys.argv[1]
    filters = fandango.first(sys.argv[2:],'')
    form = test(model,filters)

    sys.exit(app.exec_())



#!/usr/bin/env python
# Always prefer setuptools over distutils
import os, imp
from setuptools import setup, find_packages

__doc__ = """

To install as system package:

  python setup.py install
  
To install as local package, just run:

  mkdir /tmp/builds/
  python setup.py install --root=/tmp/builds
  /tmp/builds/usr/bin/$DS -? -v4

To tune some options:

  RU=/opt/control
  python setup.py egg_info --egg-base=tmp install --root=$RU/files --no-compile \
    --install-lib=lib/python/site-packages --install-scripts=bin

-------------------------------------------------------------------------------
"""

print(__doc__)

version = open('vacca/VERSION').read().strip()
scripts = []
license = 'GPL-3.0'

scripts = ['./bin/vaccagui','./bin/vaccapanel']

entry_points = {
        'console_scripts': [
            #'CopyCatDS = PyTangoArchiving.interface.CopyCatDS:main',
        ],
}
        
package_dir = {'vacca':'vacca'}
package_data = {
    'vacca': [#'VERSION','README',
         'VERSION',
         'CHANGES',
         'image/equips/*',
         'image/icons/*',
         'image/widgets/*',
         'ini/vaccagui/*',
         ]
} 
    
# PARSING EXTRA FILES ADDED TO THE PACKAGE
import os,re

if os.path.isdir('vaccagui'):
  print 'Adding custom vaccagui package ...'
  package_dir['vaccagui'] = 'vaccagui'
else:
  package_dir['vaccagui'] = 'vacca/ini/vaccagui'
  
def getter(s,d,files,remove='vaccagui'):
  #d = os.path.join(*(d.split(os.path.sep)[1:] or ['']))
  print d,files
  for n in files:
    n,r = os.path.join(d,n),re.match('.*(pyc|~|git|svn|bak|tmp)$',n)
    if os.path.isfile(n) and not r:
      if n.startswith(remove+os.path.sep):
        n = n.replace(remove+os.path.sep,'')
      s.append(n)

vg = package_dir['vaccagui']
package_data['vaccagui'] = []
os.path.walk(vg,getter,package_data['vaccagui'])
vgb = os.path.join(vg,'bin')
if os.path.isdir(vgb):
  os.path.walk(vgb,
    lambda l,d,f:l.extend(os.path.join(vgb,n) for n in f),
    scripts)
#print package_data


#import pickle
#pickle.dump(package_data,open('data.pck','w'))
  
packages = package_dir.keys()


setup(
    name="Vacca",
    version=str(version),
    license=license,
    packages=packages,
    package_dir=package_dir,
    description="Viewer and Commander Control Application for Tango Control System.",
    long_description="",
    author="Sergi Rubio",
    author_email="srubio@cells.es",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: '\
            'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
    ],
    platforms=[ "Linux" ],
    install_requires=['taurus'],
    scripts=scripts,
    entry_points=entry_points,
    include_package_data=True,
    package_data=package_data,
    zip_safe=False
  )




"""
this module will be used to document lots of things,
it parses existing modules and generate the .rst files 

"""

import vacca,inspect,fandango,os

DEFAULT_MODULE = """
%s
%s

.. automodule:: %s

"""

DEFAULT_VARIABLE = """

.. autodata:: %s.%s
   :annotation:

"""

RAW_VARIABLE = """

.. autodata:: %s.%s

"""

DEFAULT_FUNCTION = """

.. autofunction:: %s.%s

"""

DEFAULT_CLASS = """

.. autoclass:: %s.%s
   :members:

"""

TM0 = '=',True
TM1 = '=',False
TM2 = '-',False
TM3 = '~',False
TM4 = '.',False

def generate_rest_files(path='source'):
  print '\n'*5
  print 'Writing documentation settings to %s/*rst' % (path)

  submodules = [(o,v) for o,v in vars(vacca).items()
    if inspect.ismodule(v) and v.__name__.startswith('vacca')]

  for o,v in submodules:
    filename = path+'/'+o+'.rst'
    if not os.path.isfile(filename):
      print('writing %s'%filename)
      open(filename,'w').write(DEFAULT_MODULE%(v.__name__,'='*len(v.__name__),v.__name__))
      
        
  print('\nWrite this into index.rst:\n')
  print("""
  .. toctree::
     :maxdepth: 2
     
     """+
     '\n     '.join([t[0] for t in submodules]))

def get_rest_title(string,char='=',double_line=False):
    txt = char*len(string) if double_line else ''
    txt += '\n'+string+'\n'
    txt += char*len(string)
    txt += '\n\n'
    return txt

def get_vars_docs(module_name,module_vars,title='Variables',subtitle=True):
  defs = [(get_rest_title(f,*TM2) if subtitle else '')+DEFAULT_VARIABLE%(module_name,f) 
          for f in module_vars]
  if defs:
    return '\n'+get_rest_title(title,*TM1)+'\n'.join(defs)
  else:
    return ''

def get_function_docs(module_name,module_vars,title='Functions',subtitle=True):
  functions = [(f,v) for f,v in module_vars.items() 
               if inspect.isfunction(v) and v.__module__==module_name]
  defs = [(get_rest_title(f[0],*TM3) if subtitle else '')+DEFAULT_FUNCTION%(module_name,f[0]) 
          for f in functions]
  if defs:
    return '\n'+get_rest_title(title,*TM2)+'\n'.join(defs)
  else:
    return ''

def get_class_docs(module_name,module_vars,title='Classes',subtitle=True):
  classes = [(f,v) for f,v in module_vars.items() 
             if inspect.isclass(v) and v.__module__==module_name]
  defs = [(get_rest_title(f[0],*TM3) if subtitle else '')+DEFAULT_CLASS%(module_name,f[0]) 
          for f in classes]
  if defs:
    return '\n'+get_rest_title(title,*TM2)+'\n'.join(defs)
  else:
    return ''

def get_autodoc(module_name,module_scope,module_vars=[],module_doc='',module_postdoc='\n----\n'):
    module_doc = module_doc or module_scope.get('__doc__','') or ''
    #if not module_doc:
        #module_doc = get_rest_title(module_name,'=',double_line=True)
    #if ".. auto" not in m.__doc__:
    if module_vars:
        module_doc += get_vars_docs(module_name,module_vars)
    module_doc += get_class_docs(module_name,module_scope)
    module_doc += get_function_docs(module_name,module_scope)
    return module_doc+module_postdoc

if __name__ == '__main__':
  generate_rest_files()
    
__doc__ = get_autodoc(__name__,vars())
